@@ -167,6 +167,8 @@ gnome_bg_slide_show_finalize (GObject *object)
         g_queue_free_full (self->priv->stack, g_free);
 
         g_object_unref (self->priv->file);
+
+        G_OBJECT_CLASS (gnome_bg_slide_show_parent_class)->finalize (object);
 }
 
 static void
@@ -491,11 +493,7 @@ find_best_size (GSList *sizes, gint width, gint height)
 static double
 now (void)
 {
-	GTimeVal tv;
-
-	g_get_current_time (&tv);
-
-	return (double)tv.tv_sec + (tv.tv_usec / 1000000.0);
+        return (double) g_get_real_time () / 1000000.0;
 }
 
 /**
@@ -521,7 +519,7 @@ gnome_bg_slide_show_get_current_slide (GnomeBGSlideShow  *self,
                                        const char       **file1,
                                        const char       **file2)
 {
-	double delta = fmod (now() - self->priv->start_time, self->priv->total_duration);
+	double delta = fmod (now () - self->priv->start_time, self->priv->total_duration);
 	GList *list;
 	double elapsed;
 	int i;
@@ -589,7 +587,7 @@ gnome_bg_slide_show_get_slide (GnomeBGSlideShow  *self,
                                const char       **file1,
                                const char       **file2)
 {
-	double delta = fmod (now() - self->priv->start_time, self->priv->total_duration);
+	double delta = fmod (now () - self->priv->start_time, self->priv->total_duration);
         GList *l;
         int i, skipped;
         gboolean found;
