@@ -1,6 +1,7 @@
 /* gnome-rr.c
  *
  * Copyright 2007, 2008, 2013 Red Hat, Inc.
+ * Copyright 2020 NVIDIA CORPORATION
  * 
  * This file is part of the Gnome Library.
  * 
@@ -85,6 +86,7 @@ struct GnomeRROutput
     gboolean            is_presentation;
     gboolean            is_underscanning;
     gboolean            supports_underscanning;
+    gboolean            supports_color_transform;
 
     GnomeRRTile         tile_info;
 };
@@ -155,6 +157,7 @@ static gboolean gnome_rr_screen_initable_init (GInitable*, GCancellable*, GError
 static void gnome_rr_screen_initable_iface_init (GInitableIface *iface);
 static void gnome_rr_screen_async_initable_init (GAsyncInitableIface *iface);
 G_DEFINE_TYPE_WITH_CODE (GnomeRRScreen, gnome_rr_screen, G_TYPE_OBJECT,
+                         G_ADD_PRIVATE (GnomeRRScreen)
                          G_IMPLEMENT_INTERFACE (G_TYPE_INITABLE, gnome_rr_screen_initable_iface_init)
                          G_IMPLEMENT_INTERFACE (G_TYPE_ASYNC_INITABLE, gnome_rr_screen_async_initable_init))
 
@@ -873,7 +876,6 @@ void
 gnome_rr_screen_class_init (GnomeRRScreenClass *klass)
 {
     GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
-    g_type_class_add_private (klass, sizeof (GnomeRRScreenPrivate));
 
     bindtextdomain (GETTEXT_PACKAGE, GNOMELOCALEDIR);
     bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
@@ -979,8 +981,7 @@ gnome_rr_screen_class_init (GnomeRRScreenClass *klass)
 void
 gnome_rr_screen_init (GnomeRRScreen *self)
 {
-    GnomeRRScreenPrivate *priv = G_TYPE_INSTANCE_GET_PRIVATE (self, GNOME_TYPE_RR_SCREEN, GnomeRRScreenPrivate);
-    self->priv = priv;
+    self->priv = gnome_rr_screen_get_instance_private (self);
 }
 
 /* Weak reference callback set in gnome_rr_screen_new(); we remove the GObject data from the GdkScreen. */
@@ -1410,6 +1411,7 @@ output_initialize (GnomeRROutput *output, GVariant *info)
     g_variant_lookup (properties, "presentation", "b", &output->is_presentation);
     g_variant_lookup (properties, "underscanning", "b", &output->is_underscanning);
     g_variant_lookup (properties, "supports-underscanning", "b", &output->supports_underscanning);
+    g_variant_lookup (properties, "supports-color-transform", "b", &output->supports_color_transform);
 
     if ((edid = g_variant_lookup_value (properties, "edid", G_VARIANT_TYPE ("ay"))))
       {
@@ -1630,6 +1632,28 @@ gnome_rr_output_set_backlight (GnomeRROutput *output, gint value, GError **error
 								NULL, error);
 }
 
+gboolean
+gnome_rr_output_set_color_transform (GnomeRROutput *output, GnomeRRCTM ctm,
+                                     GError **error)
+{
+    GVariant *ctm_var;
+    GVariant *values[9];
+    int i;
+
+    g_return_val_if_fail (output != NULL, FALSE);
+
+    for (i = 0; i < 9; i++)
+        values[i] = g_variant_new_uint64 (ctm.matrix[i]);
+
+    ctm_var = g_variant_new_tuple (values, 9);
+
+    return meta_dbus_display_config_call_set_output_ctm_sync (output->info->screen->priv->proxy,
+                                                              output->info->serial,
+                                                              output->id,
+                                                              ctm_var,
+                                                              NULL, error);
+}
+
 /**
  * gnome_rr_screen_get_output_by_name:
  *
@@ -2245,6 +2269,13 @@ gnome_rr_output_supports_underscanning (GnomeRROutput *output)
     return output->supports_underscanning;
 }
 
+gboolean
+gnome_rr_output_supports_color_transform (const GnomeRROutput *output)
+{
+    g_assert (output != NULL);
+    return output->supports_color_transform;
+}
+
 const char *
 _gnome_rr_output_get_connector_type (GnomeRROutput *output)
 {
